// https://github.com/Fil/attitude v0.2.5 Copyright 2022 Philippe Rivière
(function (global, factory) {
typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
typeof define === 'function' && define.amd ? define(factory) :
(global = typeof globalThis !== 'undefined' ? globalThis : global || self, global.attitude = factory());
})(this, (function () { 'use strict';

const abs = Math.abs;
const atan = Math.atan;
const atan2 = Math.atan2;
const cos = Math.cos;
const floor = Math.floor;
const hypot = Math.hypot;
const sign = Math.sign;
const sin = Math.sin;
const sqrt = Math.sqrt;
const tan = Math.tan;

const pi = Math.PI;
const halfPi = pi / 2;

const degrees = 180 / pi;
const radians = 1 / degrees;

function acos(x) {
  return x > 1 ? 0 : x < -1 ? pi : Math.acos(x);
}
function asin(x) {
  return x > 1 ? halfPi : x < -1 ? -halfPi : Math.asin(x);
}

// https://observablehq.com/@fil/sinpi-cospi

// adapted from https://github.com/math-io/sinpi
function sinpi(x) {
  if (!isFinite(x)) return NaN;
  const r = x % 2,
    s = sign(r),
    a = abs(r);
  return a == 0 || a == 1
    ? 0 * s
    : a < .25
    ? sin(pi * r)
    : a < .75
    ? cos(pi * (.5 - a)) * s
    : a < 1.25
    ? sin(pi * (s - r))
    : a < 1.75
    ? -s * cos(pi * (a - 1.5))
    : sin(pi * (r - 2 * s));
}


// adapted from https://github.com/math-io/cospi
function cospi(x) {
  if (!isFinite(x)) return NaN;
  const a = abs(x),
    i = floor(a),
    r = a - i;
  if (r == .5) return 0;
  return (
    (i % 2 == 1 ? -1 : 1) *
    (r < .25 ? cos(pi * r) : r < .75 ? sin(pi * (.5 - r)) : -cos(pi * (1 - r)))
  );
}

function cartesiand([l, p]) {
  const cp = cospi(p / 180);
  return [cp * cospi(l / 180), cp * sinpi(l / 180), sinpi(p / 180)];
}

function sphericald([x, y, z]) {
  return [atan2(y, x) * degrees, asin(z) * degrees];
}

function dot(a, b) {
  return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

function cross(a, b) {
  return [
    a[1] * b[2] - a[2] * b[1],
    a[2] * b[0] - a[0] * b[2],
    a[0] * b[1] - a[1] * b[0]
  ];
}

const versor_normalize = q => {
  const n = hypot(...q);
  return q.map(d => d / n);
};

// eq. (198) & (199)
// - axis as (lon, lat) in degrees
// - angle in degrees
// much more precision for small angles!
function versor_toAxisAngle(q) {
  const s = sqrt(
      abs(q[0]) < 0.5
        ? 1 - q[0] * q[0]
        : q[1] * q[1] + q[2] * q[2] + q[3] * q[3]
    ),
    a =
      abs(q[0]) < 0.5
        ? 2 * acos(q[0])
        : q[0] < 0
        ? 2 * pi - 2 * asin(s)
        : 2 * asin(s);
  if (!s) return { axis: [0, 90], angle: 0 };
  const s1 = 1 / s;
  return {
    axis: sphericald([q[3] * s1, -q[2] * s1, q[1] * s1]),
    angle: a * degrees
  };
}

// eq. (175)
// - axis as (lon, lat) in degrees
// - angle in degrees
function versor_fromAxisAngle(axis, angle) {
  const c = cospi(angle / 360),
    s = sinpi(angle / 360),
    d = cartesiand(axis);
  return [c, d[2] * s, -d[1] * s, d[0] * s];
}

// Euler 123 eq (297)
function versor_fromEulerAngles([l, p, g]) {

  // fix south pole
  if (p + 90 < 1e-9) p = -90 + 1e-9;
  
  const sl = sinpi(l / 360),
    cl = cospi(l / 360),
    sp = sinpi(p / 360),
    cp = cospi(p / 360),
    sg = sinpi((g||0) / 360),
    cg = cospi((g||0) / 360);
  return [
    cl * cp * cg + sl * sp * sg,
    sl * cp * cg - cl * sp * sg,
    cl * sp * cg + sl * cp * sg,
    cl * cp * sg - sl * sp * cg
  ];
}

// eq (290)
function versor_toEulerAngles([q0, q1, q2, q3]) {
  const a = 2 * q2 * q3 + 2 * q0 * q1,
    b = q3 * q3 - q1 * q1 + q0 * q0 - q2 * q2,
    c = 2 * q1 * q3 - 2 * q0 * q2,
    d = 2 * q1 * q2 + 2 * q0 * q3,
    e = -q3 * q3 - q2 * q2 + q1 * q1 + q0 * q0,
    l = atan2(a, b),
    g = atan2(d, e);

  // fix north pole
  const eps = 1e-9;
  if (
    c < 0 &&
    ((abs(a) < eps && abs(b) < eps) || (abs(d) < eps && abs(e) < eps))
  ) {
    const app = versor_toEulerAngles([q0, q2, q1, q3]);
    return [app[1] / 2, app[0], -app[1] / 2];
  }

  return [l * degrees, -asin(c) * degrees, g * degrees];
}

// https://observablehq.com/@d3/world-tour
function versor_multiply([a1, b1, c1, d1], [a2, b2, c2, d2]) {
  return [
    a1 * a2 - b1 * b2 - c1 * c2 - d1 * d2,
    a1 * b2 + b1 * a2 + c1 * d2 - d1 * c2,
    a1 * c2 - b1 * d2 + c1 * a2 + d1 * b2,
    a1 * d2 + b1 * c2 - c1 * b2 + d1 * a2
  ];
}

function transpose([[r11, r12, r13], [r21, r22, r23], [r31, r32, r33]]) {
  return [[r11, r21, r31], [r12, r22, r32], [r13, r23, r33]];
}

// eq. (5) (we use z = R^T(z’))
function MatrixRotatePoint(
  [[r11, r12, r13], [r21, r22, r23], [r31, r32, r33]],
  point
) {
  function rotator(point) {
    const [x, y, z] = cartesiand(point);
    const product = [
      r11 * x + r21 * y + r31 * z,
      r12 * x + r22 * y + r32 * z,
      r13 * x + r23 * y + r33 * z
    ];
    return sphericald(product);
  }
  return point ? rotator(point) : rotator;
}

// eq (289)
function RotationMatrix_toEulerAngles([
  [r11 /* r12 */ /* r13 */, ,],
  [r21 /* r22 */ /* r23 */, ,],
  [r31, r32, r33]
]) {
  const [f, t, p] = [atan2(r32, r33), -asin(r31), atan2(r21, r11)],
    [lambda, phi, gamma] = [-p * degrees, t * degrees, -f * degrees];

  return [lambda, phi, gamma];
}

// eq (184)
// changes for precision by fil@rezo.net
function RotationMatrix_fromAxisAngle(axis, angle) {
  const [n1, n2, n3] = cartesiand(axis),
    [n12, n22, n32] = [n1 * n1, n2 * n2, n3 * n3],
    c = cospi(angle / 180),
    s2 = (1 - c) / 2, // cos(a/2)^2
    c2 = (1 + c) / 2, // sin(a/2)^2
    sc = sinpi(angle / 180) / 2; // sin(a/2)cos(a/2)

  return [
    [
      (n12 - n32 - n22) * s2 + c2,
      2 * n1 * n2 * s2 + 2 * n3 * sc,
      2 * n1 * n3 * s2 - 2 * n2 * sc
    ],
    [
      2 * n1 * n2 * s2 - 2 * n3 * sc,
      (n22 - n32 - n12) * s2 + c2,
      2 * n2 * n3 * s2 + 2 * n1 * sc
    ],
    [
      2 * n1 * n3 * s2 + 2 * n2 * sc,
      2 * n2 * n3 * s2 - 2 * n1 * sc,
      (n32 - n22 - n12) * s2 + c2
    ]
  ];
}

const vectorEquidistant = {
  forward: a => a,
  inverse: a => a
};

const vectorStereographic = {
  forward: a => tan(a / 4),
  inverse: a => 4 * atan(a)
};

const vectorGnomonic = {
  forward: a => tan(a / 2),
  inverse: a => 2 * atan(a)
};

function attitude(init = {}) {
  let axis, angle, q, matrix;

  function rotate(point) {
    return point === undefined ? rotate : MatrixRotatePoint(matrix, point);
  }

  function set_axis_angle(x, a) {
    axis = x;
    angle = a;
    q = versor_fromAxisAngle(axis, angle);
    matrix = RotationMatrix_fromAxisAngle(axis, angle);
    return rotate;
  }

  function set_versor(p) {
    const a = versor_toAxisAngle((q = p));
    matrix = RotationMatrix_fromAxisAngle((axis = a.axis), (angle = a.angle));
    return rotate;
  }

  function get_vector(f) {
    const n = f(angle * radians),
      v = cartesiand(axis);
    return v.map(d => d * n);
  }
  function set_vector(v, f_1) {
    const n = sqrt(dot(v, v)),
      axis = n > 0
        ? sphericald(v.map(d => d / n))
        : [0, 0];
    set_axis_angle(axis, f_1(n) * degrees);
    return rotate;
  }

  rotate.axis = _ => (_ === undefined ? axis : set_axis_angle(_, angle));
  rotate.angle = _ => (_ === undefined ? angle : set_axis_angle(axis, +_));
  rotate.versor = _ => (_ === undefined ? q : set_versor(versor_normalize(_)));

  rotate.vectorGnomonic = _ =>
    _ === undefined
      ? get_vector(vectorGnomonic.forward)
      : set_vector(_, vectorGnomonic.inverse);
  rotate.vectorStereographic = _ =>
    _ === undefined
      ? get_vector(vectorStereographic.forward)
      : set_vector(_, vectorStereographic.inverse);
  rotate.vectorEquidistant = _ =>
    _ === undefined
      ? get_vector(vectorEquidistant.forward)
      : set_vector(_, vectorEquidistant.inverse);
  rotate.vector = rotate.vectorStereographic;

  rotate.matrix = _ =>
    _ === undefined
      ? matrix
      : set_versor(
          versor_normalize(
            versor_fromEulerAngles(RotationMatrix_toEulerAngles(_))
          )
        );

  rotate.angles = _ =>
    _ === undefined
      ? versor_toEulerAngles(q)
      : set_versor(versor_fromEulerAngles(_));

  rotate.invert = point => MatrixRotatePoint(transpose(matrix), point);

  // returns a new attitude with inverse parameters
  rotate.inverse = () => attitude({ axis, angle: -angle });

  // returns a new product attitude
  rotate.compose = _ => attitude().versor(versor_multiply(q, _.versor()));

  // returns a new power attitude
  rotate.power = _ => attitude({ axis, angle: angle * +_ });

  // returns a new attitude that rotates along the arc of great circle (A,B)
  rotate.arc = (A, B) => attitude(arc(A, B));

  // returns an interpolator between two attitudes
  rotate.interpolateTo = b => interpolateAttitude(rotate, b);

  // accept d3.geoRotation's init with Euler Angles
  if (init.length >= 2) {
    rotate.angles(init);
  } else {
    set_axis_angle(
      (axis = init.axis || [0, 90]),
      (angle = init.angle === undefined ? 360 : +init.angle)
    );
  }

  return rotate;
}

// arc(A, B) returns an axis-angle that rotates A to B along a great circle (shortest path)
function arc(A, B) {
  const a = cartesiand(A),
    b = cartesiand(B);
  const alpha = acos(dot(a, b)), // angle
    w = cross(a, b), // axis direction
    n = sqrt(dot(w, w)); // axis norm
  return !n
    ? { angle: dot(a, b) > 0 ? 0 : 180 }
    : {
        axis: sphericald([w[0] / n, w[1] / n, w[2] / n]),
        angle: alpha * degrees
      };
}

function interpolateAttitude(a, b) {
  const c = b.compose(a.inverse());
  return t => (t === 1 ? b : !t ? a : c.power(t).compose(a));
}

return attitude;

}));
